from cpl_query.extension import List

from py_to_uxf_core.abc.umlet_creator_abc import UmletCreatorABC
from py_to_uxf_core.model.class_implementation import ClassImplementation
from py_to_uxf_core.model.dimension import Dimension
from py_to_uxf_core.model.position import Position
from py_to_uxf_core.model.python_class import PythonClass
from py_to_uxf_core.model.uml_class import UMLClass


class UmletCreatorService(UmletCreatorABC):

    def __init__(self):
        UmletCreatorABC.__init__(self)

        self._padding = 30

    def _sort_by_implementation(self, uml_classes: List[UMLClass], implementations: List[ClassImplementation]):
        base_sub_map: dict[UMLClass, list[UMLClass]] = {}
        moved_classes: list[UMLClass] = []
        for implementation in implementations:
            sub_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.subclass).first()
            base_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.base).first()

            if base_class not in base_sub_map:
                base_sub_map[base_class] = []

            base_sub_map[base_class].append(sub_class)

        last_base = Position(self._padding, self._padding)
        for base in base_sub_map:
            subclasses = base_sub_map[base]

            base.position.x = last_base.x
            base.position.y = self._padding

            last_sub = Position(base.position.x, base.position.y + base.dimension.height)
            for sub_class in subclasses:
                sub_class.position.x = base.position.x
                sub_class.position.y = base.position.y

                if sub_class.position.x <= last_sub.x:
                    sub_class.position.x = last_sub.x
                if sub_class.position.y <= last_sub.y:
                    sub_class.position.y = last_sub.y + self._padding * 2

                last_sub.x = sub_class.position.x + sub_class.dimension.width + self._padding
                moved_classes.append(sub_class)

            last_base = last_sub
            last_base.x += self._padding * 2
            moved_classes.append(base)

        for cls in uml_classes:
            if cls in moved_classes:
                continue

            cls.position.x = last_base.x + self._padding
            last_base.x = cls.position.x + cls.dimension.width

    def generate_xml(self, classes: List[PythonClass], implementations: List[ClassImplementation]) -> str:
        uml_classes = List(UMLClass)
        xml_classes = ''
        xml_relations = ''
        default_width = 80
        default_height = 80
        next_x = self._padding

        for cls in classes:
            uml_cls = UMLClass(cls, Position(next_x, self._padding), Dimension(default_width, default_height))
            # trigger xml generation to render coordinates
            uml_cls.create_xml_body()
            next_x += round(uml_cls.dimension.width, -1) + self._padding
            uml_classes.append(uml_cls)

        self._sort_by_implementation(uml_classes, implementations)
        for implementation in implementations:
            sub_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.subclass).first()
            base_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.base).first()
            xml_relations += sub_class.implementation_as_xml(base_class)

        for uml_cls in uml_classes:
            # save class xml
            xml_classes += uml_cls.as_xml()

        return f"""\
        <?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <diagram program="umlet" version="14.1.1">
            <zoom_level>10</zoom_level>
            {xml_classes}
            {xml_relations}
        </diagram>
        """.replace('    ', '').replace('\t', '')
